from cfinterface.components.section import Section
from typing import IO, List, Dict, Union, Callable, Any
import numpy as np  # type: ignore
import numpy.typing as npt
import pandas as pd  # type: ignore
from enum import Enum


class VariavelOperacao(Enum):
    MERCADO = "MER"
    ENERGIA_ARMAZENADA_ABSOLUTA_INICIAL = "EARMI"
    ENERGIA_NATURAL_AFLUENTE_ABSOLUTA = "ENAA"
    GERACAO_HIDRAULICA_CONTROLAVEL = "GHIDC"
    ENERGIA_VERTIDA = "EVER"
    ENERGIA_ARMAZENADA_ABSOLUTA_FINAL = "EARMF"
    ENERGIA_NATURAL_AFLUENTE_FIO_BRUTA = "ENAFB"
    ENERGIA_EVAPORADA = "EVAP"
    ENERGIA_ENCHIMENTO_VOLUME_MORTO = "EVMOR"
    GERACAO_TERMICA = "GTER"
    DEFICIT = "DEF"
    VALOR_AGUA = "VAGUA"
    CUSTO_MARGINAL_OPERACAO = "CMO"
    GERACAO_HIDRAULICA_FIO_LIQUIDA = "GHIDFL"
    PERDAS_GERACAO_HIDRAULICA_FIO = "PGHIDF"
    INTERCAMBIO = "INT"
    EXCESSO = "EXC"
    ENERGIA_NATURAL_AFLUENTE_BRUTA = "ENAB"
    ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL_CORRIGIDA = "ENAC"
    GERACAO_HIDRAULICA_MAXIMA = "GHIDMAX"
    ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO = "ENACD"
    ENERGIA_AFLUENTE_FIO_DESVIO = "ENAFD"
    BENEFICIO_INTERCAMBIO = "BINT"
    FATOR_CORRECAO_ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL = "FCENA"
    VIOLACAO_CURVA_AVERSAO = "VVMINOP"
    ACIONAMENTO_CURVA_AVERSAO = "AVMINOP"
    PENALIDADE_CURVA_AVERSAO = "PVMINOP"
    CUSTO_OPERACAO = "COP"
    CUSTO_GERACAO_TERMICA = "CTER"
    BENEFICIO_AGRUPAMENTO_INTERCAMBIO = "BAGRINT"
    ENERGIA_NATURAL_AFLUENTE_FIO = "ENAF"
    BENEFICIO_DESPACHO_GNL = "BGNL"
    VIOLACAO_GERACAO_HIRAULICA_MINIMA = "VGHIDMIN"
    VIOLACAO_ENERGIA_VAZAO_MINIMA = "VEVMIN"
    INVASAO_SAR = "ISAR"
    ACIONAMENTO_SAR = "ASAR"
    PENALIDADE_SAR = "PSAR"
    GERACAO_HIDRAULICA_MAXIMA_CONSIDERANDO_RE = "GHIDMAXR"
    VOLUME_ARMAZENADO_ABSOLUTO_FINAL = "VARMF"
    GERACAO_HIDRAULICA_USINA = "GHIDU"
    VOLUME_TURBINADO = "VTUR"
    VOLUME_VERTIDO = "VVER"
    VIOLACAO_GERACAO_HIDRAULICA_MINIMA_USINA = "VGHIDMINU"
    ENCHIMENTO_VOLUME_MORTO_USINA = "VMORU"
    VIOLACAO_DEFLUENCIA_MINIMA = "VDEFMIN"
    VOLUME_DESVIO_USINA = "VOLDU"
    VOLUME_DESVIO_POSITIVO_USINA = "VOLDPU"
    VOLUME_DESVIO_NEGATIVO_USINA = "VOLDNU"
    ##
    VIOLACAO_FPHA = "VFPHA"
    VAZAO_AFLUENTE = "QAFL"
    VAZAO_INCREMENTAL = "QINC"
    VOLUME_ARMAZENADO_PERCENTUAL_FINAL = "VARPF"
    GEOL_GSOL_OLD = "GEOL_GSOL_OLD"
    VIOLACAO_GEE = "VIOLACAO_GEE"
    CUSTO_VIOLACAO_ENERGIA_VAZAO_MINIMA = "CVEVMIN"
    CUSTO_ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO = "CENACD"
    CUSTO_ENERGIA_AFLUENTE_FIO_DESVIO = "CENAFD"
    CUSTO_VIOLACAO_GERACAO_HIDRAULICA_MINIMA = "CVGHIDMIN"
    SOMA_ENERGIA_NATURAL_AFLUENTE_12_MESES = "ENA12"
    SOMA_VAZAO_AFLUENTE_12_MESES = "QAFL12"
    GERACAO_EOLICA = "GEOL"
    VELOCIDADE_VENTO = "VENTO"
    VIOLACAO_FUNCAO_PRODUCAO_EOLICA = "VFPEOL"
    VIOLACAO_DEFLUENCIA_MAXIMA = "VDEFMAX"
    VIOLACAO_TURBINAMENTO_MINIMO = "VTURMIN"
    VIOLACAO_TURBINAMENTO_MAXIMO = "VTURMAX"
    VIOLACAO_LPP_TURBINAMENTO_MAXIMO = "VLPPTURMAX"
    VIOLACAO_LPP_DEFLUENCIA_MAXIMA = "VLPPDEFMAX"
    VIOLACAO_LPP_TURBINAMENTO_MAXIMO_USINA = "VLPPTURMAXU"
    VIOLACAO_LPP_DEFLUENCIA_MAXIMA_USINA = "VLPPDEFMAXU"
    RHS_LPP_TURBINAMENTO_MAXIMO = "RHSLPPTURMAX"
    RHS_LPP_DEFLUENCIA_MAXIMA = "RHSLPPDEFMAX"
    RHS_LPP_TURBINAMENTO_MAXIMO_USINA = "RHSLPPTURMAXU"
    RHS_LPP_DEFLUENCIA_MAXIMA_USINA = "RHSLPPDEFMAXU"
    VIOLACAO_RESTRICOES_ELETRICAS_ESPECIAIS = "VREESP"
    CUSTO_RESTRICOES_ELETRICAS_ESPECIAIS = "CREESP"
    VOLUME_ARMAZENADO_ABSOLUTO_INICIAL = "VARMI"
    VALOR_AGUA_USINA = "VAGUAU"
    VOLUME_EVAPORADO = "VEVAP"
    VOLUME_BOMBEADO = "VBOMB"
    CONSUMO_ENERGIA_ESTACAO_BOMBEAMENTO = "EBOMB"
    VOLUME_CANAL_DESVIO_USINA = "VOLCDU"


class SecaoDadosForward(Section):
    """
    Registro com os resultados da operação do modelo NEWAVE.
    """

    VARIAVEIS: List[VariavelOperacao] = [
        VariavelOperacao.MERCADO,
        VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_INICIAL,
        VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_ABSOLUTA,
        VariavelOperacao.GERACAO_HIDRAULICA_CONTROLAVEL,
        VariavelOperacao.ENERGIA_VERTIDA,
        VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_FINAL,
        VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO_BRUTA,
        VariavelOperacao.ENERGIA_EVAPORADA,
        VariavelOperacao.ENERGIA_ENCHIMENTO_VOLUME_MORTO,
        VariavelOperacao.GERACAO_TERMICA,
        VariavelOperacao.DEFICIT,
        VariavelOperacao.VALOR_AGUA,
        VariavelOperacao.CUSTO_MARGINAL_OPERACAO,
        VariavelOperacao.GERACAO_HIDRAULICA_FIO_LIQUIDA,
        VariavelOperacao.PERDAS_GERACAO_HIDRAULICA_FIO,
        VariavelOperacao.INTERCAMBIO,
        VariavelOperacao.EXCESSO,
        VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_BRUTA,
        VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL_CORRIGIDA,
        VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA,
        VariavelOperacao.ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO,
        VariavelOperacao.ENERGIA_AFLUENTE_FIO_DESVIO,
        VariavelOperacao.BENEFICIO_INTERCAMBIO,
        VariavelOperacao.FATOR_CORRECAO_ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL,
        VariavelOperacao.VIOLACAO_CURVA_AVERSAO,
        VariavelOperacao.ACIONAMENTO_CURVA_AVERSAO,
        VariavelOperacao.PENALIDADE_CURVA_AVERSAO,
        VariavelOperacao.CUSTO_OPERACAO,
        VariavelOperacao.CUSTO_GERACAO_TERMICA,
        VariavelOperacao.BENEFICIO_AGRUPAMENTO_INTERCAMBIO,
        VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO,
        VariavelOperacao.BENEFICIO_DESPACHO_GNL,
        VariavelOperacao.VIOLACAO_GERACAO_HIRAULICA_MINIMA,
        VariavelOperacao.VIOLACAO_ENERGIA_VAZAO_MINIMA,
        VariavelOperacao.INVASAO_SAR,
        VariavelOperacao.ACIONAMENTO_SAR,
        VariavelOperacao.PENALIDADE_SAR,
        VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA_CONSIDERANDO_RE,
        VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_FINAL,
        VariavelOperacao.GERACAO_HIDRAULICA_USINA,
        VariavelOperacao.VOLUME_TURBINADO,
        VariavelOperacao.VOLUME_VERTIDO,
        VariavelOperacao.VIOLACAO_GERACAO_HIDRAULICA_MINIMA_USINA,
        VariavelOperacao.ENCHIMENTO_VOLUME_MORTO_USINA,
        VariavelOperacao.VIOLACAO_DEFLUENCIA_MINIMA,
        VariavelOperacao.VOLUME_DESVIO_USINA,
        VariavelOperacao.VOLUME_DESVIO_POSITIVO_USINA,
        VariavelOperacao.VOLUME_DESVIO_NEGATIVO_USINA,
        VariavelOperacao.VIOLACAO_FPHA,
        VariavelOperacao.VAZAO_AFLUENTE,
        VariavelOperacao.VAZAO_INCREMENTAL,
        VariavelOperacao.VOLUME_ARMAZENADO_PERCENTUAL_FINAL,
        VariavelOperacao.GEOL_GSOL_OLD,
        VariavelOperacao.VIOLACAO_GEE,
        VariavelOperacao.CUSTO_VIOLACAO_ENERGIA_VAZAO_MINIMA,
        VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO,
        VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_FIO_DESVIO,
        VariavelOperacao.CUSTO_VIOLACAO_GERACAO_HIDRAULICA_MINIMA,
        VariavelOperacao.SOMA_ENERGIA_NATURAL_AFLUENTE_12_MESES,
        VariavelOperacao.SOMA_VAZAO_AFLUENTE_12_MESES,
        VariavelOperacao.GERACAO_EOLICA,
        VariavelOperacao.VELOCIDADE_VENTO,
        VariavelOperacao.VIOLACAO_FUNCAO_PRODUCAO_EOLICA,
        VariavelOperacao.VIOLACAO_DEFLUENCIA_MAXIMA,
        VariavelOperacao.VIOLACAO_TURBINAMENTO_MAXIMO,
        VariavelOperacao.VIOLACAO_TURBINAMENTO_MINIMO,
        VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO,
        VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA,
        VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO_USINA,
        VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA_USINA,
        VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO,
        VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA,
        VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO_USINA,
        VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA_USINA,
        VariavelOperacao.VIOLACAO_RESTRICOES_ELETRICAS_ESPECIAIS,
        VariavelOperacao.CUSTO_RESTRICOES_ELETRICAS_ESPECIAIS,
        VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_INICIAL,
        VariavelOperacao.VALOR_AGUA_USINA,
        VariavelOperacao.VOLUME_EVAPORADO,
        VariavelOperacao.VOLUME_BOMBEADO,
        VariavelOperacao.CONSUMO_ENERGIA_ESTACAO_BOMBEAMENTO,
        VariavelOperacao.VOLUME_CANAL_DESVIO_USINA,
    ]

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.data: Dict[VariavelOperacao, Union[np.ndarray, pd.DataFrame]] = {}

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, SecaoDadosForward):
            return False
        bloco: SecaoDadosForward = o
        if not all(
            [
                isinstance(self.data, dict),
                isinstance(bloco.data, dict),
            ]
        ):
            return False
        else:
            if not all(
                [isinstance(b, pd.DataFrame) for b in self.data.values()]
                + [isinstance(b, pd.DataFrame) for b in o.data.values()]
            ):
                return False
            else:
                return all(
                    [
                        (k1 == k2) and v1.equals(v2)  # type: ignore
                        for (k1, v1), (k2, v2) in zip(
                            self.data.items(), bloco.data.items()
                        )
                    ]
                )

    def __le_e_atribui_int(
        self, file: IO, destino: np.ndarray, tamanho: int, indice: int
    ):
        offset = tamanho * indice
        destino[offset : offset + tamanho] = np.frombuffer(
            file.read(tamanho * 4),
            dtype=np.int32,
            count=tamanho,
        )

    def __le_e_atribui_float(
        self, file: IO, destino: np.ndarray, tamanho: int, indice: int
    ):
        offset = tamanho * indice
        destino[offset : offset + tamanho] = np.frombuffer(
            file.read(tamanho * 4),
            dtype=np.float32,
            count=tamanho,
        )

    def __inicializa_variaveis(
        self, numero_estagios: int, numero_forwards: int
    ):
        # Variáveis que sempre existem
        self.__estagios_df = np.repeat(
            np.arange(1, numero_estagios + 1), numero_forwards
        )
        self.__cenarios_df = np.tile(
            np.arange(1, numero_forwards + 1), numero_estagios
        )
        self.__num_simulacoes = numero_estagios * numero_forwards
        self.estagio = np.zeros((1 * self.__num_simulacoes,), dtype=np.int32)
        # Variáveis que variam com a versão do modelo
        mapa_variaveis: Dict[
            VariavelOperacao, Callable[[Any], npt.NDArray[np.float32]]
        ] = {
            VariavelOperacao.MERCADO: lambda _: np.zeros(
                (self.numero_submercados * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_INICIAL: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,), dtype=np.float32
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_ABSOLUTA: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,), dtype=np.float32
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_CONTROLAVEL: lambda _: np.zeros(
                (
                    self.numero_rees
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.ENERGIA_VERTIDA: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,), dtype=np.float32
            ),
            VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_FINAL: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,), dtype=np.float32
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO_BRUTA: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,), dtype=np.float32
            ),
            VariavelOperacao.ENERGIA_EVAPORADA: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,), dtype=np.float32
            ),
            VariavelOperacao.ENERGIA_ENCHIMENTO_VOLUME_MORTO: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,), dtype=np.float32
            ),
            VariavelOperacao.GERACAO_TERMICA: lambda _: np.zeros(
                (
                    2
                    * self.total_classes_termicas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.DEFICIT: lambda _: np.zeros(
                (
                    self.numero_submercados
                    * self.numero_patamares_deficit
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VALOR_AGUA: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.CUSTO_MARGINAL_OPERACAO: lambda _: np.zeros(
                (
                    self.numero_submercados
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_FIO_LIQUIDA: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.PERDAS_GERACAO_HIDRAULICA_FIO: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.INTERCAMBIO: lambda _: np.zeros(
                (
                    self.numero_total_submercados
                    * (self.numero_total_submercados - 1)
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.EXCESSO: lambda _: np.zeros(
                (
                    self.numero_submercados
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_BRUTA: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL_CORRIGIDA: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA: lambda _: np.zeros(
                (
                    self.numero_rees
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.ENERGIA_AFLUENTE_FIO_DESVIO: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.BENEFICIO_INTERCAMBIO: lambda _: np.zeros(
                (
                    self.numero_total_submercados
                    * (self.numero_total_submercados - 1)
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.FATOR_CORRECAO_ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL: lambda _: np.zeros(  # noqa
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_CURVA_AVERSAO: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.ACIONAMENTO_CURVA_AVERSAO: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.PENALIDADE_CURVA_AVERSAO: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.CUSTO_OPERACAO: lambda _: np.zeros(
                (1 * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.CUSTO_GERACAO_TERMICA: lambda _: np.zeros(
                (self.numero_submercados * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.BENEFICIO_AGRUPAMENTO_INTERCAMBIO: lambda _: np.zeros(
                (
                    self.numero_patamares_carga
                    * self.numero_agrupamentos_intercambio
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.BENEFICIO_DESPACHO_GNL: lambda _: np.zeros(
                (
                    self.numero_submercados
                    * self.numero_patamares_carga
                    * self.lag_maximo_usinas_gnl
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_GERACAO_HIRAULICA_MINIMA: lambda _: np.zeros(
                (
                    self.numero_rees
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_ENERGIA_VAZAO_MINIMA: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.INVASAO_SAR: lambda _: np.zeros(
                (1 * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.ACIONAMENTO_SAR: lambda _: np.zeros(
                (1 * self.__num_simulacoes,),
                dtype=np.int32,
            ),
            VariavelOperacao.PENALIDADE_SAR: lambda _: np.zeros(
                (1 * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA_CONSIDERANDO_RE: lambda _: np.zeros(
                (
                    self.numero_rees
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_FINAL: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_USINA: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VOLUME_TURBINADO: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VOLUME_VERTIDO: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_GERACAO_HIDRAULICA_MINIMA_USINA: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.ENCHIMENTO_VOLUME_MORTO_USINA: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_DEFLUENCIA_MINIMA: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VOLUME_DESVIO_USINA: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.VOLUME_DESVIO_POSITIVO_USINA: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.VOLUME_DESVIO_NEGATIVO_USINA: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_FPHA: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VAZAO_AFLUENTE: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.VAZAO_INCREMENTAL: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.GEOL_GSOL_OLD: lambda _: np.zeros(
                (
                    2
                    * self.numero_submercados
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_GEE: lambda _: np.zeros(
                (1 * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.VOLUME_ARMAZENADO_PERCENTUAL_FINAL: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.CUSTO_VIOLACAO_ENERGIA_VAZAO_MINIMA: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_FIO_DESVIO: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.CUSTO_VIOLACAO_GERACAO_HIDRAULICA_MINIMA: lambda _: np.zeros(
                (
                    self.numero_rees
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.SOMA_ENERGIA_NATURAL_AFLUENTE_12_MESES: lambda _: np.zeros(
                (self.numero_rees * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.SOMA_VAZAO_AFLUENTE_12_MESES: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.GERACAO_EOLICA: lambda _: np.zeros(
                (
                    self.numero_parques_eolicos_equivalentes
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VELOCIDADE_VENTO: lambda _: np.zeros(
                (
                    self.numero_parques_eolicos_equivalentes
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_FUNCAO_PRODUCAO_EOLICA: lambda _: np.zeros(
                (
                    self.numero_parques_eolicos_equivalentes
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_DEFLUENCIA_MAXIMA: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_TURBINAMENTO_MINIMO: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_TURBINAMENTO_MAXIMO: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO: lambda _: np.zeros(
                (
                    self.numero_rees
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA: lambda _: np.zeros(
                (
                    self.numero_rees
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO_USINA: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA_USINA: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO: lambda _: np.zeros(
                (
                    self.numero_rees
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA: lambda _: np.zeros(
                (
                    self.numero_rees
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO_USINA: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA_USINA: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VIOLACAO_RESTRICOES_ELETRICAS_ESPECIAIS: lambda _: np.zeros(
                (
                    self.numero_restricoes_eletricas_especiais
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.CUSTO_RESTRICOES_ELETRICAS_ESPECIAIS: lambda _: np.zeros(
                (
                    self.numero_restricoes_eletricas_especiais
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_INICIAL: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.VALOR_AGUA_USINA: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.VOLUME_EVAPORADO: lambda _: np.zeros(
                (self.numero_usinas_hidreletricas * self.__num_simulacoes,),
                dtype=np.float32,
            ),
            VariavelOperacao.VOLUME_BOMBEADO: lambda _: np.zeros(
                (
                    self.numero_estacoes_bombeamento
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.CONSUMO_ENERGIA_ESTACAO_BOMBEAMENTO: lambda _: np.zeros(
                (
                    self.numero_estacoes_bombeamento
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
            VariavelOperacao.VOLUME_CANAL_DESVIO_USINA: lambda _: np.zeros(
                (
                    self.numero_usinas_hidreletricas
                    * self.numero_patamares_carga
                    * self.__num_simulacoes,
                ),
                dtype=np.float32,
            ),
        }
        for v in self.__class__.VARIAVEIS:
            self.data[v] = mapa_variaveis[v](v)

    def __le_registro(
        self,
        file: IO,
        offset: int,
        indice: int,
    ):
        file.seek(offset)
        self.__le_e_atribui_int(file, self.estagio, 1, indice)
        # Variáveis que variam com a versão do modelo
        mapa_variaveis: Dict[VariavelOperacao, Callable] = {
            VariavelOperacao.MERCADO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.MERCADO],
                self.numero_submercados,
                indice,
            ),
            VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_INICIAL: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_INICIAL
                ],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_ABSOLUTA: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_ABSOLUTA],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_CONTROLAVEL: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.GERACAO_HIDRAULICA_CONTROLAVEL],
                self.numero_rees * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.ENERGIA_VERTIDA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.ENERGIA_VERTIDA],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_FINAL: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_FINAL],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO_BRUTA: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO_BRUTA],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.ENERGIA_EVAPORADA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.ENERGIA_EVAPORADA],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.ENERGIA_ENCHIMENTO_VOLUME_MORTO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.ENERGIA_ENCHIMENTO_VOLUME_MORTO],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.GERACAO_TERMICA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.GERACAO_TERMICA],
                2 * self.total_classes_termicas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.DEFICIT: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.DEFICIT],
                self.numero_submercados
                * self.numero_patamares_deficit
                * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VALOR_AGUA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VALOR_AGUA],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.CUSTO_MARGINAL_OPERACAO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.CUSTO_MARGINAL_OPERACAO],
                self.numero_submercados * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_FIO_LIQUIDA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.GERACAO_HIDRAULICA_FIO_LIQUIDA],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.PERDAS_GERACAO_HIDRAULICA_FIO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.PERDAS_GERACAO_HIDRAULICA_FIO],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.INTERCAMBIO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.INTERCAMBIO],
                self.numero_total_submercados
                * (self.numero_total_submercados - 1)
                * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.EXCESSO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.EXCESSO],
                self.numero_submercados * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_BRUTA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_BRUTA],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL_CORRIGIDA: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL_CORRIGIDA
                ],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA],
                self.numero_rees * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO
                ],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.ENERGIA_AFLUENTE_FIO_DESVIO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.ENERGIA_AFLUENTE_FIO_DESVIO],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.BENEFICIO_INTERCAMBIO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.BENEFICIO_INTERCAMBIO],
                self.numero_total_submercados
                * (self.numero_total_submercados - 1)
                * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.FATOR_CORRECAO_ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.FATOR_CORRECAO_ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL
                ],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.VIOLACAO_CURVA_AVERSAO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_CURVA_AVERSAO],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.ACIONAMENTO_CURVA_AVERSAO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.ACIONAMENTO_CURVA_AVERSAO],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.PENALIDADE_CURVA_AVERSAO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.PENALIDADE_CURVA_AVERSAO],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.CUSTO_OPERACAO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.CUSTO_OPERACAO],
                1,
                indice,
            ),
            VariavelOperacao.CUSTO_GERACAO_TERMICA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.CUSTO_GERACAO_TERMICA],
                self.numero_submercados,
                indice,
            ),
            VariavelOperacao.BENEFICIO_AGRUPAMENTO_INTERCAMBIO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.BENEFICIO_AGRUPAMENTO_INTERCAMBIO],
                self.numero_patamares_carga
                * self.numero_agrupamentos_intercambio,
                indice,
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.BENEFICIO_DESPACHO_GNL: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.BENEFICIO_DESPACHO_GNL],
                self.numero_submercados
                * self.numero_patamares_carga
                * self.lag_maximo_usinas_gnl,
                indice,
            ),
            VariavelOperacao.VIOLACAO_GERACAO_HIRAULICA_MINIMA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_GERACAO_HIRAULICA_MINIMA],
                self.numero_rees * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VIOLACAO_ENERGIA_VAZAO_MINIMA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_ENERGIA_VAZAO_MINIMA],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.INVASAO_SAR: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.INVASAO_SAR],
                1,
                indice,
            ),
            VariavelOperacao.ACIONAMENTO_SAR: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.ACIONAMENTO_SAR],
                1,
                indice,
            ),
            VariavelOperacao.PENALIDADE_SAR: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.PENALIDADE_SAR],
                1,
                indice,
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA_CONSIDERANDO_RE: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA_CONSIDERANDO_RE
                ],
                self.numero_rees * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_FINAL: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_FINAL],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_USINA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.GERACAO_HIDRAULICA_USINA],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VOLUME_TURBINADO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VOLUME_TURBINADO],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VOLUME_VERTIDO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VOLUME_VERTIDO],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VIOLACAO_GERACAO_HIDRAULICA_MINIMA_USINA: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.VIOLACAO_GERACAO_HIDRAULICA_MINIMA_USINA
                ],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.ENCHIMENTO_VOLUME_MORTO_USINA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.ENCHIMENTO_VOLUME_MORTO_USINA],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.VIOLACAO_DEFLUENCIA_MINIMA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_DEFLUENCIA_MINIMA],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VOLUME_DESVIO_USINA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VOLUME_DESVIO_USINA],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.VOLUME_DESVIO_POSITIVO_USINA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VOLUME_DESVIO_POSITIVO_USINA],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.VOLUME_DESVIO_NEGATIVO_USINA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VOLUME_DESVIO_NEGATIVO_USINA],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.VIOLACAO_FPHA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_FPHA],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VAZAO_AFLUENTE: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VAZAO_AFLUENTE],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.VAZAO_INCREMENTAL: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VAZAO_INCREMENTAL],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.VOLUME_ARMAZENADO_PERCENTUAL_FINAL: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[VariavelOperacao.VOLUME_ARMAZENADO_PERCENTUAL_FINAL],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.GEOL_GSOL_OLD: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.GEOL_GSOL_OLD],
                2 * self.numero_submercados * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VIOLACAO_GEE: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_GEE],
                1,
                indice,
            ),
            VariavelOperacao.CUSTO_VIOLACAO_ENERGIA_VAZAO_MINIMA: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.CUSTO_VIOLACAO_ENERGIA_VAZAO_MINIMA
                ],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO
                ],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_FIO_DESVIO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_FIO_DESVIO],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.CUSTO_VIOLACAO_GERACAO_HIDRAULICA_MINIMA: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.CUSTO_VIOLACAO_GERACAO_HIDRAULICA_MINIMA
                ],
                self.numero_rees * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.SOMA_ENERGIA_NATURAL_AFLUENTE_12_MESES: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.SOMA_ENERGIA_NATURAL_AFLUENTE_12_MESES
                ],
                self.numero_rees,
                indice,
            ),
            VariavelOperacao.SOMA_VAZAO_AFLUENTE_12_MESES: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.SOMA_VAZAO_AFLUENTE_12_MESES],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.GERACAO_EOLICA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.GERACAO_EOLICA],
                self.numero_parques_eolicos_equivalentes
                * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VELOCIDADE_VENTO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VELOCIDADE_VENTO],
                self.numero_parques_eolicos_equivalentes,
                indice,
            ),
            VariavelOperacao.VIOLACAO_FUNCAO_PRODUCAO_EOLICA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_FUNCAO_PRODUCAO_EOLICA],
                self.numero_parques_eolicos_equivalentes
                * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VIOLACAO_DEFLUENCIA_MAXIMA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_DEFLUENCIA_MAXIMA],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VIOLACAO_TURBINAMENTO_MAXIMO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_TURBINAMENTO_MAXIMO],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VIOLACAO_TURBINAMENTO_MINIMO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_TURBINAMENTO_MINIMO],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO],
                self.numero_rees * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA],
                self.numero_rees * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO_USINA: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO_USINA
                ],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA_USINA: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA_USINA
                ],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO],
                self.numero_rees * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA],
                self.numero_rees * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO_USINA: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO_USINA],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA_USINA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA_USINA],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VIOLACAO_RESTRICOES_ELETRICAS_ESPECIAIS: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.VIOLACAO_RESTRICOES_ELETRICAS_ESPECIAIS
                ],
                self.numero_restricoes_eletricas_especiais
                * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.CUSTO_RESTRICOES_ELETRICAS_ESPECIAIS: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.CUSTO_RESTRICOES_ELETRICAS_ESPECIAIS
                ],
                self.numero_restricoes_eletricas_especiais
                * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_INICIAL: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_INICIAL],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.VALOR_AGUA_USINA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VALOR_AGUA_USINA],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.VOLUME_EVAPORADO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VOLUME_EVAPORADO],
                self.numero_usinas_hidreletricas,
                indice,
            ),
            VariavelOperacao.VOLUME_BOMBEADO: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VOLUME_BOMBEADO],
                self.numero_estacoes_bombeamento * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.CONSUMO_ENERGIA_ESTACAO_BOMBEAMENTO: lambda _: self.__le_e_atribui_float(  # noqa
                file,
                self.data[
                    VariavelOperacao.CONSUMO_ENERGIA_ESTACAO_BOMBEAMENTO
                ],
                self.numero_estacoes_bombeamento * self.numero_patamares_carga,
                indice,
            ),
            VariavelOperacao.VOLUME_CANAL_DESVIO_USINA: lambda _: self.__le_e_atribui_float(
                file,
                self.data[VariavelOperacao.VOLUME_CANAL_DESVIO_USINA],
                self.numero_usinas_hidreletricas * self.numero_patamares_carga,
                indice,
            ),
        }
        for v in self.__class__.VARIAVEIS:
            mapa_variaveis[v](v)

    def __converte_array_em_dataframe(
        self,
        variavel: np.ndarray,
        colunas_identificacao: dict,
        num_elementos: int,
    ):
        return pd.DataFrame(
            data={
                **{
                    "estagio": np.repeat(self.__estagios_df, num_elementos),
                    "cenario": np.repeat(self.__cenarios_df, num_elementos),
                },
                **colunas_identificacao,
                **{"valor": variavel},
            }
        )

    def __converte_array_submercado(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "submercado": np.tile(
                    self.nomes_submercados, self.__num_simulacoes
                )
            },
            self.numero_submercados,
        )

    def __converte_array_submercado_patamar(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "submercado": np.tile(
                    np.repeat(
                        self.nomes_submercados, self.numero_patamares_carga
                    ),
                    self.__num_simulacoes,
                ),
                "patamar": np.tile(
                    np.tile(
                        np.arange(1, self.numero_patamares_carga + 1),
                        self.numero_submercados,
                    ),
                    self.__num_simulacoes,
                ),
            },
            self.numero_submercados * self.numero_patamares_carga,
        )

    def __converte_array_ree(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {"ree": np.tile(self.nomes_rees, self.__num_simulacoes)},
            self.numero_rees,
        )

    def __converte_array_ree_patamar(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "ree": np.tile(
                    np.repeat(
                        self.nomes_rees,
                        self.numero_patamares_carga,
                    ),
                    self.__num_simulacoes,
                ),
                "patamar": np.tile(
                    np.tile(
                        np.arange(1, self.numero_patamares_carga + 1),
                        self.numero_rees,
                    ),
                    self.__num_simulacoes,
                ),
            },
            self.numero_rees * self.numero_patamares_carga,
        )

    def __converte_array_ute_patamar(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "usina": np.tile(
                    np.tile(
                        self.nomes_classes_termicas,
                        self.numero_patamares_carga,
                    ),
                    self.__num_simulacoes,
                ),
                "patamar": np.tile(
                    np.repeat(
                        np.arange(1, self.numero_patamares_carga + 1),
                        self.total_classes_termicas,
                    ),
                    self.__num_simulacoes,
                ),
            },
            self.total_classes_termicas * self.numero_patamares_carga,
        )

    def __converte_array_submercado_patamardeficit_patamar(
        self, variavel: np.ndarray
    ):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "submercado": np.tile(
                    np.repeat(
                        self.nomes_submercados,
                        self.numero_patamares_deficit
                        * self.numero_patamares_carga,
                    ),
                    self.__num_simulacoes,
                ),
                "patamarDeficit": np.tile(
                    np.tile(
                        np.arange(1, self.numero_patamares_deficit + 1),
                        self.numero_patamares_carga * self.numero_submercados,
                    ),
                    self.__num_simulacoes,
                ),
                "patamar": np.tile(
                    np.tile(
                        np.repeat(
                            np.arange(1, self.numero_patamares_carga + 1),
                            self.numero_patamares_deficit,
                        ),
                        self.numero_submercados,
                    ),
                    self.__num_simulacoes,
                ),
            },
            self.numero_submercados
            * self.numero_patamares_deficit
            * self.numero_patamares_carga,
        )

    def __converte_array_par_submercados_patamar(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "submercadoDe": np.tile(
                    np.tile(
                        self.nomes_submercados_totais[:-1],
                        self.numero_total_submercados
                        * self.numero_patamares_carga,
                    ),
                    self.__num_simulacoes,
                ),
                "submercadoPara": np.tile(
                    np.tile(
                        np.repeat(
                            self.nomes_submercados_totais,
                            self.numero_total_submercados - 1,
                        ),
                        self.numero_patamares_carga,
                    ),
                    self.__num_simulacoes,
                ),
                "patamar": np.tile(
                    np.tile(
                        np.arange(1, self.numero_patamares_carga + 1),
                        self.numero_total_submercados
                        * (self.numero_total_submercados - 1),
                    ),
                    self.__num_simulacoes,
                ),
            },
            self.numero_total_submercados
            * (self.numero_total_submercados - 1)
            * self.numero_patamares_carga,
        )

    def __converte_array_agrupamentointercambio_patamar(
        self, variavel: np.ndarray
    ):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "agrupamentoIntercambio": np.tile(
                    np.repeat(
                        np.arange(1, self.numero_agrupamentos_intercambio + 1),
                        self.numero_patamares_carga,
                    ),
                    self.__num_simulacoes,
                ),
                "patamar": np.tile(
                    np.arange(1, self.numero_patamares_carga + 1),
                    self.numero_agrupamentos_intercambio
                    * self.__num_simulacoes,
                ),
            },
            self.numero_agrupamentos_intercambio * self.numero_patamares_carga,
        )

    def __converte_array_laggnl_submercado_patamar(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "lag": np.tile(
                    np.arange(1, self.lag_maximo_usinas_gnl + 1),
                    self.numero_patamares_carga
                    * self.numero_submercados
                    * self.__num_simulacoes,
                ),
                "patamar": np.tile(
                    np.repeat(
                        np.arange(1, self.numero_patamares_carga + 1),
                        self.lag_maximo_usinas_gnl,
                    ),
                    self.numero_submercados * self.__num_simulacoes,
                ),
                "submercado": np.tile(
                    np.repeat(
                        self.nomes_submercados,
                        self.lag_maximo_usinas_gnl
                        * self.numero_patamares_carga,
                    ),
                    self.__num_simulacoes,
                ),
            },
            self.lag_maximo_usinas_gnl
            * self.numero_submercados
            * self.numero_patamares_carga,
        )

    def __converte_array_uhe(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "usina": np.tile(
                    self.nomes_usinas_hidreletricas, self.__num_simulacoes
                )
            },
            self.numero_usinas_hidreletricas,
        )

    def __converte_array_uhe_patamar(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "usina": np.tile(
                    np.repeat(
                        self.nomes_usinas_hidreletricas,
                        self.numero_patamares_carga,
                    ),
                    self.__num_simulacoes,
                ),
                "patamar": np.tile(
                    np.tile(
                        np.arange(1, self.numero_patamares_carga + 1),
                        self.numero_usinas_hidreletricas,
                    ),
                    self.__num_simulacoes,
                ),
            },
            self.numero_usinas_hidreletricas * self.numero_patamares_carga,
        )

    def __converte_array_pee(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "pee": np.tile(
                    self.nomes_parques_eolicos_equivalentes,
                    self.__num_simulacoes,
                )
            },
            self.numero_parques_eolicos_equivalentes,
        )

    def __converte_array_pee_patamar(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "pee": np.tile(
                    np.repeat(
                        self.nomes_parques_eolicos_equivalentes,
                        self.numero_patamares_carga,
                    ),
                    self.__num_simulacoes,
                ),
                "patamar": np.tile(
                    np.tile(
                        np.arange(1, self.numero_patamares_carga + 1),
                        self.numero_parques_eolicos_equivalentes,
                    ),
                    self.__num_simulacoes,
                ),
            },
            self.numero_parques_eolicos_equivalentes
            * self.numero_patamares_carga,
        )

    def __converte_array_restricaoeletrica_patamar(self, variavel: np.ndarray):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "restricao": np.tile(
                    np.repeat(
                        np.arange(
                            1,
                            self.numero_restricoes_eletricas_especiais + 1,
                        ),
                        self.numero_patamares_carga,
                    ),
                    self.__num_simulacoes,
                ),
                "patamar": np.tile(
                    np.tile(
                        np.arange(1, self.numero_patamares_carga + 1),
                        self.numero_restricoes_eletricas_especiais,
                    ),
                    self.__num_simulacoes,
                ),
            },
            self.numero_restricoes_eletricas_especiais
            * self.numero_patamares_carga,
        )

    def __converte_array_estacaobombeamento_patamar(
        self, variavel: np.ndarray
    ):
        return self.__converte_array_em_dataframe(
            variavel,
            {
                "estacao": np.tile(
                    np.repeat(
                        self.nomes_estacoes_bombeamento,
                        self.numero_patamares_carga,
                    ),
                    self.__num_simulacoes,
                ),
                "patamar": np.tile(
                    np.tile(
                        np.arange(1, self.numero_patamares_carga + 1),
                        self.numero_estacoes_bombeamento,
                    ),
                    self.__num_simulacoes,
                ),
            },
            self.numero_estacoes_bombeamento * self.numero_patamares_carga,
        )

    def __converte_arrays_em_dataframes(self):
        mapa_variaveis: Dict[VariavelOperacao, Callable] = {
            VariavelOperacao.MERCADO: lambda _: self.__converte_array_submercado(
                self.data[VariavelOperacao.MERCADO]
            ),
            VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_INICIAL: lambda _: self.__converte_array_ree(  # noqa
                self.data[VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_INICIAL]
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_ABSOLUTA: lambda _: self.__converte_array_ree(  # noqa
                self.data[VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_ABSOLUTA],
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_CONTROLAVEL: lambda _: self.__converte_array_ree_patamar(  # noqa
                self.data[VariavelOperacao.GERACAO_HIDRAULICA_CONTROLAVEL]
            ),
            VariavelOperacao.ENERGIA_VERTIDA: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.ENERGIA_VERTIDA]
            ),
            VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_FINAL: lambda _: self.__converte_array_ree(  # noqa
                self.data[VariavelOperacao.ENERGIA_ARMAZENADA_ABSOLUTA_FINAL]
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO_BRUTA: lambda _: self.__converte_array_ree(  # noqa
                self.data[VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO_BRUTA]
            ),
            VariavelOperacao.ENERGIA_EVAPORADA: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.ENERGIA_EVAPORADA]
            ),
            VariavelOperacao.ENERGIA_ENCHIMENTO_VOLUME_MORTO: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.ENERGIA_ENCHIMENTO_VOLUME_MORTO]
            ),
            VariavelOperacao.GERACAO_TERMICA: lambda _: self.__converte_array_ute_patamar(
                self.data[VariavelOperacao.GERACAO_TERMICA][::2]
                + self.data[VariavelOperacao.GERACAO_TERMICA][1::2]
            ),
            VariavelOperacao.DEFICIT: lambda _: self.__converte_array_submercado_patamardeficit_patamar(  # noqa
                self.data[VariavelOperacao.DEFICIT]
            ),
            VariavelOperacao.VALOR_AGUA: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.VALOR_AGUA]
            ),
            VariavelOperacao.CUSTO_MARGINAL_OPERACAO: lambda _: self.__converte_array_submercado_patamar(  # noqa
                self.data[VariavelOperacao.CUSTO_MARGINAL_OPERACAO]
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_FIO_LIQUIDA: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.GERACAO_HIDRAULICA_FIO_LIQUIDA]
            ),
            VariavelOperacao.PERDAS_GERACAO_HIDRAULICA_FIO: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.PERDAS_GERACAO_HIDRAULICA_FIO]
            ),
            VariavelOperacao.INTERCAMBIO: lambda _: self.__converte_array_par_submercados_patamar(
                self.data[VariavelOperacao.INTERCAMBIO]
            ),
            VariavelOperacao.EXCESSO: lambda _: self.__converte_array_submercado_patamar(
                self.data[VariavelOperacao.EXCESSO]
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_BRUTA: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_BRUTA]
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL_CORRIGIDA: lambda _: self.__converte_array_ree(  # noqa
                self.data[
                    VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL_CORRIGIDA
                ]
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA: lambda _: self.__converte_array_ree_patamar(
                self.data[VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA]
            ),
            VariavelOperacao.ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO: lambda _: self.__converte_array_ree(  # noqa
                self.data[VariavelOperacao.ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO]
            ),
            VariavelOperacao.ENERGIA_AFLUENTE_FIO_DESVIO: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.ENERGIA_AFLUENTE_FIO_DESVIO]
            ),
            VariavelOperacao.BENEFICIO_INTERCAMBIO: lambda _: self.__converte_array_par_submercados_patamar(  # noqa
                self.data[VariavelOperacao.BENEFICIO_INTERCAMBIO]
            ),
            VariavelOperacao.FATOR_CORRECAO_ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL: lambda _: self.__converte_array_ree(  # noqa
                self.data[
                    VariavelOperacao.FATOR_CORRECAO_ENERGIA_NATURAL_AFLUENTE_CONTROLAVEL
                ]
            ),
            VariavelOperacao.VIOLACAO_CURVA_AVERSAO: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.VIOLACAO_CURVA_AVERSAO]
            ),
            VariavelOperacao.ACIONAMENTO_CURVA_AVERSAO: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.ACIONAMENTO_CURVA_AVERSAO]
            ),
            VariavelOperacao.PENALIDADE_CURVA_AVERSAO: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.PENALIDADE_CURVA_AVERSAO]
            ),
            VariavelOperacao.CUSTO_OPERACAO: lambda _: self.__converte_array_em_dataframe(
                self.data[VariavelOperacao.CUSTO_OPERACAO],
                {},
                1,
            ),
            VariavelOperacao.CUSTO_GERACAO_TERMICA: lambda _: self.__converte_array_submercado(
                self.data[VariavelOperacao.CUSTO_GERACAO_TERMICA]
            ),
            VariavelOperacao.BENEFICIO_AGRUPAMENTO_INTERCAMBIO: lambda _: self.__converte_array_agrupamentointercambio_patamar(  # noqa
                self.data[VariavelOperacao.BENEFICIO_AGRUPAMENTO_INTERCAMBIO]
            ),
            VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.ENERGIA_NATURAL_AFLUENTE_FIO]
            ),
            VariavelOperacao.BENEFICIO_DESPACHO_GNL: lambda _: self.__converte_array_laggnl_submercado_patamar(  # noqa
                self.data[VariavelOperacao.BENEFICIO_DESPACHO_GNL]
            ),
            VariavelOperacao.VIOLACAO_GERACAO_HIRAULICA_MINIMA: lambda _: self.__converte_array_ree_patamar(  # noqa
                self.data[VariavelOperacao.VIOLACAO_GERACAO_HIRAULICA_MINIMA]
            ),
            VariavelOperacao.VIOLACAO_ENERGIA_VAZAO_MINIMA: lambda _: self.__converte_array_ree(
                self.data[VariavelOperacao.VIOLACAO_ENERGIA_VAZAO_MINIMA]
            ),
            VariavelOperacao.INVASAO_SAR: lambda _: self.__converte_array_em_dataframe(
                self.data[VariavelOperacao.INVASAO_SAR],
                {},
                1,
            ),
            VariavelOperacao.ACIONAMENTO_SAR: lambda _: self.__converte_array_em_dataframe(
                self.data[VariavelOperacao.ACIONAMENTO_SAR],
                {},
                1,
            ),
            VariavelOperacao.PENALIDADE_SAR: lambda _: self.__converte_array_em_dataframe(
                self.data[VariavelOperacao.PENALIDADE_SAR],
                {},
                1,
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA_CONSIDERANDO_RE: lambda _: self.__converte_array_ree_patamar(  # noqa
                self.data[
                    VariavelOperacao.GERACAO_HIDRAULICA_MAXIMA_CONSIDERANDO_RE
                ]
            ),
            VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_FINAL: lambda _: self.__converte_array_uhe(  # noqa
                self.data[VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_FINAL]
            ),
            VariavelOperacao.GERACAO_HIDRAULICA_USINA: lambda _: self.__converte_array_uhe_patamar(  # noqa
                self.data[VariavelOperacao.GERACAO_HIDRAULICA_USINA]
            ),
            VariavelOperacao.VOLUME_TURBINADO: lambda _: self.__converte_array_uhe_patamar(
                self.data[VariavelOperacao.VOLUME_TURBINADO]
            ),
            VariavelOperacao.VOLUME_VERTIDO: lambda _: self.__converte_array_uhe_patamar(
                self.data[VariavelOperacao.VOLUME_VERTIDO]
            ),
            VariavelOperacao.VIOLACAO_GERACAO_HIDRAULICA_MINIMA_USINA: lambda _: self.__converte_array_uhe_patamar(  # noqa
                self.data[
                    VariavelOperacao.VIOLACAO_GERACAO_HIDRAULICA_MINIMA_USINA
                ]
            ),
            VariavelOperacao.ENCHIMENTO_VOLUME_MORTO_USINA: lambda _: self.__converte_array_uhe(
                self.data[VariavelOperacao.ENCHIMENTO_VOLUME_MORTO_USINA]
            ),
            VariavelOperacao.VIOLACAO_DEFLUENCIA_MINIMA: lambda _: self.__converte_array_uhe_patamar(  # noqa
                self.data[VariavelOperacao.VIOLACAO_DEFLUENCIA_MINIMA]
            ),
            VariavelOperacao.VOLUME_DESVIO_USINA: lambda _: self.__converte_array_uhe(
                self.data[VariavelOperacao.VOLUME_DESVIO_USINA]
            ),
            VariavelOperacao.VOLUME_DESVIO_POSITIVO_USINA: lambda _: self.__converte_array_uhe(
                self.data[VariavelOperacao.VOLUME_DESVIO_POSITIVO_USINA]
            ),
            VariavelOperacao.VOLUME_DESVIO_NEGATIVO_USINA: lambda _: self.__converte_array_uhe(
                self.data[VariavelOperacao.VOLUME_DESVIO_NEGATIVO_USINA]
            ),
            VariavelOperacao.VIOLACAO_FPHA: lambda _: self.__converte_array_uhe_patamar(
                self.data[VariavelOperacao.VIOLACAO_FPHA]
            ),
            VariavelOperacao.VAZAO_AFLUENTE: lambda _: self.__converte_array_uhe(
                self.data[VariavelOperacao.VAZAO_AFLUENTE]
            ),
            VariavelOperacao.VAZAO_INCREMENTAL: lambda _: self.__converte_array_uhe(
                self.data[VariavelOperacao.VAZAO_INCREMENTAL]
            ),
            VariavelOperacao.VOLUME_ARMAZENADO_PERCENTUAL_FINAL: lambda _: self.__converte_array_uhe(  # noqa
                self.data[VariavelOperacao.VOLUME_ARMAZENADO_PERCENTUAL_FINAL]
            ),
            VariavelOperacao.GEOL_GSOL_OLD: lambda _: pd.DataFrame(),
            VariavelOperacao.VIOLACAO_GEE: lambda _: pd.DataFrame(),
            VariavelOperacao.CUSTO_VIOLACAO_ENERGIA_VAZAO_MINIMA: lambda _: self.__converte_array_ree(  # noqa
                self.data[VariavelOperacao.CUSTO_VIOLACAO_ENERGIA_VAZAO_MINIMA]
            ),
            VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO: lambda _: self.__converte_array_ree(  # noqa
                self.data[
                    VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_CONTROLAVEL_DESVIO
                ]
            ),
            VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_FIO_DESVIO: lambda _: self.__converte_array_ree(  # noqa
                self.data[VariavelOperacao.CUSTO_ENERGIA_AFLUENTE_FIO_DESVIO]
            ),
            VariavelOperacao.CUSTO_VIOLACAO_GERACAO_HIDRAULICA_MINIMA: lambda _: self.__converte_array_ree_patamar(  # noqa
                self.data[
                    VariavelOperacao.CUSTO_VIOLACAO_GERACAO_HIDRAULICA_MINIMA
                ]
            ),
            VariavelOperacao.SOMA_ENERGIA_NATURAL_AFLUENTE_12_MESES: lambda _: self.__converte_array_ree(  # noqa
                self.data[
                    VariavelOperacao.SOMA_ENERGIA_NATURAL_AFLUENTE_12_MESES
                ]
            ),
            VariavelOperacao.SOMA_VAZAO_AFLUENTE_12_MESES: lambda _: self.__converte_array_uhe(
                self.data[VariavelOperacao.SOMA_VAZAO_AFLUENTE_12_MESES]
            ),
            VariavelOperacao.GERACAO_EOLICA: lambda _: self.__converte_array_pee_patamar(
                self.data[VariavelOperacao.GERACAO_EOLICA]
            ),
            VariavelOperacao.VELOCIDADE_VENTO: lambda _: self.__converte_array_pee(
                self.data[VariavelOperacao.VELOCIDADE_VENTO]
            ),
            VariavelOperacao.VIOLACAO_FUNCAO_PRODUCAO_EOLICA: lambda _: self.__converte_array_pee_patamar(  # noqa
                self.data[VariavelOperacao.VIOLACAO_FUNCAO_PRODUCAO_EOLICA]
            ),
            VariavelOperacao.VIOLACAO_DEFLUENCIA_MAXIMA: lambda _: self.__converte_array_uhe_patamar(  # noqa
                self.data[VariavelOperacao.VIOLACAO_DEFLUENCIA_MAXIMA]
            ),
            VariavelOperacao.VIOLACAO_TURBINAMENTO_MAXIMO: lambda _: self.__converte_array_uhe_patamar(  # noqa
                self.data[VariavelOperacao.VIOLACAO_TURBINAMENTO_MAXIMO]
            ),
            VariavelOperacao.VIOLACAO_TURBINAMENTO_MINIMO: lambda _: self.__converte_array_uhe_patamar(  # noqa
                self.data[VariavelOperacao.VIOLACAO_TURBINAMENTO_MINIMO]
            ),
            VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO: lambda _: self.__converte_array_ree_patamar(  # noqa
                self.data[VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO]
            ),
            VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA: lambda _: self.__converte_array_ree_patamar(  # noqa
                self.data[VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA]
            ),
            VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO_USINA: lambda _: self.__converte_array_uhe_patamar(  # noqa
                self.data[
                    VariavelOperacao.VIOLACAO_LPP_TURBINAMENTO_MAXIMO_USINA
                ]
            ),
            VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA_USINA: lambda _: self.__converte_array_uhe_patamar(  # noqa
                self.data[
                    VariavelOperacao.VIOLACAO_LPP_DEFLUENCIA_MAXIMA_USINA
                ]
            ),
            VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO: lambda _: self.__converte_array_ree_patamar(  # noqa
                self.data[VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO]
            ),
            VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA: lambda _: self.__converte_array_ree_patamar(  # noqa
                self.data[VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA]
            ),
            VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO_USINA: lambda _: self.__converte_array_uhe_patamar(  # noqa
                self.data[VariavelOperacao.RHS_LPP_TURBINAMENTO_MAXIMO_USINA]
            ),
            VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA_USINA: lambda _: self.__converte_array_uhe_patamar(  # noqa
                self.data[VariavelOperacao.RHS_LPP_DEFLUENCIA_MAXIMA_USINA]
            ),
            VariavelOperacao.VIOLACAO_RESTRICOES_ELETRICAS_ESPECIAIS: lambda _: self.__converte_array_restricaoeletrica_patamar(  # noqa
                self.data[
                    VariavelOperacao.VIOLACAO_RESTRICOES_ELETRICAS_ESPECIAIS
                ]
            ),
            VariavelOperacao.CUSTO_RESTRICOES_ELETRICAS_ESPECIAIS: lambda _: self.__converte_array_restricaoeletrica_patamar(  # noqa
                self.data[
                    VariavelOperacao.CUSTO_RESTRICOES_ELETRICAS_ESPECIAIS
                ]
            ),
            VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_INICIAL: lambda _: self.__converte_array_uhe(  # noqa
                self.data[VariavelOperacao.VOLUME_ARMAZENADO_ABSOLUTO_INICIAL]
            ),
            VariavelOperacao.VALOR_AGUA_USINA: lambda _: self.__converte_array_uhe(
                self.data[VariavelOperacao.VALOR_AGUA_USINA]
            ),
            VariavelOperacao.VOLUME_EVAPORADO: lambda _: self.__converte_array_uhe(
                self.data[VariavelOperacao.VOLUME_EVAPORADO]
            ),
            VariavelOperacao.VOLUME_BOMBEADO: lambda _: self.__converte_array_estacaobombeamento_patamar(  # noqa
                self.data[VariavelOperacao.VOLUME_BOMBEADO]
            ),
            VariavelOperacao.CONSUMO_ENERGIA_ESTACAO_BOMBEAMENTO: lambda _: self.__converte_array_estacaobombeamento_patamar(  # noqa
                self.data[VariavelOperacao.CONSUMO_ENERGIA_ESTACAO_BOMBEAMENTO]
            ),
            VariavelOperacao.VOLUME_CANAL_DESVIO_USINA: lambda _: self.__converte_array_uhe_patamar(  # noqa
                self.data[VariavelOperacao.VOLUME_CANAL_DESVIO_USINA]
            ),
        }
        for v in self.__class__.VARIAVEIS:
            self.data[v] = mapa_variaveis[v](v)

        # Soma GTMIN e GTFLEX
        # TODO - talvez tenha uma diferença aqui, pois o NEWAVE
        # escreve por submercado. Para casos de PMO, as térmicas são
        # cadastradas por submercado no conft, então não deve
        # dar diferença. Os nomes deveriam ser fornecidos por
        # submercado já..

    def read(
        self,
        file: IO,
        tamanho_registro: int = 41264,
        numero_estagios: int = 60,
        numero_forwards: int = 200,
        numero_patamares_carga: int = 3,
        numero_patamares_deficit: int = 1,
        numero_agrupamentos_intercambio: int = 1,
        numero_classes_termicas_submercados: List[int] = [],
        lag_maximo_usinas_gnl: int = 2,
        numero_restricoes_eletricas_especiais: int = 0,
        nomes_submercados: List[str] = ["SUDESTE", "SUL", "NORDESTE", "NORTE"],
        nomes_submercados_totais: List[str] = [
            "SUDESTE",
            "SUL",
            "NORDESTE",
            "NORTE",
            "NOFICT1",
        ],
        nomes_rees: List[str] = [
            "SUDESTE",
            "MADEIRA",
            "TPIRES",
            "ITAIPU",
            "PARANA",
            "PRNPANEMA",
            "SUL",
            "IGUACU",
            "NORDESTE",
            "NORTE",
            "BMONTE",
            "MAN-AP",
        ],
        nomes_classes_termicas: List[str] = [],
        nomes_usinas_hidreletricas: List[str] = [],
        nomes_parques_eolicos_equivalentes: List[str] = [],
        nomes_estacoes_bombeamento: List[str] = [],
        *args,
        **kwargs,
    ):
        # Atribui variáveis locais
        self.numero_rees = len(nomes_rees)
        self.numero_submercados = len(nomes_submercados)
        self.numero_total_submercados = len(nomes_submercados_totais)
        self.numero_patamares_carga = numero_patamares_carga
        self.numero_patamares_deficit = numero_patamares_deficit
        self.numero_agrupamentos_intercambio = numero_agrupamentos_intercambio
        self.numero_classes_termicas_submercados = (
            numero_classes_termicas_submercados
        )
        self.total_classes_termicas = sum(
            self.numero_classes_termicas_submercados
        )
        self.numero_usinas_hidreletricas = len(nomes_usinas_hidreletricas)
        self.lag_maximo_usinas_gnl = lag_maximo_usinas_gnl
        self.numero_parques_eolicos_equivalentes = len(
            nomes_parques_eolicos_equivalentes
        )
        self.numero_restricoes_eletricas_especiais = (
            numero_restricoes_eletricas_especiais
        )
        self.numero_estacoes_bombeamento = len(nomes_estacoes_bombeamento)
        self.nomes_submercados = np.array(nomes_submercados)
        self.nomes_submercados_totais = np.array(nomes_submercados_totais)
        self.nomes_rees = np.array(nomes_rees)
        self.nomes_classes_termicas = np.array(nomes_classes_termicas)
        self.nomes_usinas_hidreletricas = np.array(nomes_usinas_hidreletricas)
        self.nomes_parques_eolicos_equivalentes = np.array(
            nomes_parques_eolicos_equivalentes
        )
        self.nomes_estacoes_bombeamento = np.array(nomes_estacoes_bombeamento)
        # Realiza leitura
        self.__inicializa_variaveis(numero_estagios, numero_forwards)
        for estagio in range(numero_estagios):
            for serie in range(numero_forwards):
                indice = estagio * numero_forwards + serie
                offset = tamanho_registro * indice
                self.__le_registro(file, offset, indice)
        self.__converte_arrays_em_dataframes()



from inewave.newave.modelos.engnat import SecaoDadosEngnat
from inewave.newave.engnat import Engnat
from inewave.config import MAX_ANOS_HISTORICO

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch


ARQ_TESTE = "./tests/mocks/arquivos/engnat.dat"

NUM_CONFIGURACOES = 1
NUM_REES = 12
ANO_INICIO_HISTORICO = 1931
NUM_ENTRADAS = NUM_CONFIGURACOES * NUM_REES * 12 * MAX_ANOS_HISTORICO


def test_secao_engnat():
    r = SecaoDadosEngnat()
    with open(ARQ_TESTE, "rb") as fp:
        r.read(
            fp,
            numero_rees=NUM_REES,
            ano_inicio_historico=ANO_INICIO_HISTORICO,
            numero_configuracoes=NUM_CONFIGURACOES,
        )

    assert len(r.data) == NUM_ENTRADAS


def test_atributos_encontrados_engnat():
    h = Engnat.read(
        ARQ_TESTE,
        numero_rees=NUM_REES,
        ano_inicio_historico=ANO_INICIO_HISTORICO,
        numero_configuracoes=NUM_CONFIGURACOES,
    )
    assert h.series is not None
    assert h.series.isna().sum().sum() == 0


def test_atributos_nao_encontrados_engnat():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        h = Engnat.read(
            ARQ_TESTE,
            numero_rees=NUM_REES,
            ano_inicio_historico=ANO_INICIO_HISTORICO,
            numero_configuracoes=NUM_CONFIGURACOES,
        )
        assert h.series.isna().sum().sum() == NUM_ENTRADAS


def test_eq_engnat():
    h1 = Engnat.read(
        ARQ_TESTE,
        numero_rees=NUM_REES,
        ano_inicio_historico=ANO_INICIO_HISTORICO,
        numero_configuracoes=NUM_CONFIGURACOES,
    )
    h2 = Engnat.read(
        ARQ_TESTE,
        numero_rees=NUM_REES,
        ano_inicio_historico=ANO_INICIO_HISTORICO,
        numero_configuracoes=NUM_CONFIGURACOES,
    )
    assert h1 == h2



from inewave.newave.modelos.blocos.versaomodelo import VersaoModelo
from inewave.newave.modelos.avl_cortesfpha_nwv import TabelaAvlCortesFpha

from inewave.newave.modelos.arquivoscsv.arquivocsv import ArquivoCSV
from typing import Optional
import pandas as pd  # type: ignore


class AvlCortesFpha(ArquivoCSV):
    """
    Arquivo com os cortes da função de produção para as UHEs
    do NEWAVE.
    """

    BLOCKS = [VersaoModelo, TabelaAvlCortesFpha]

    @property
    def tabela(self) -> Optional[pd.DataFrame]:
        """
        A tabela de dados que está contida no arquivo.

        - codigo_usina (`int`)
        - periodo (`int`)
        - nome_usina (`str`)
        - indice_corte (`int`)
        - fator_correcao (`float`)
        - rhs_energia (`float`)
        - coeficiente_volume_util_MW_hm3 (`float`)
        - coeficiente_vazao_turbinada_MW_m3s (`float`)
        - coeficiente_vazao_vertida_MW_m3s (`float`)
        - coeficiente_vazao_lateral_MW_m3s (`float`)

        :return: A tabela como um dataframe
        :rtype: pd.DataFrame | None
        """
        return self._tabela()



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousinapatamar import (
    ArquivoUsinaPatamar,
)
from inewave.nwlistop.modelos.dtbmin import DtbminAnos


class Dtbmin(ArquivoUsinaPatamar):
    """
    Armazena os dados das saídas referentes à violação de restrição de
    turbinamento mínimo por usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `dtbmin00x.out`, onde x varia conforme
    a usina em questão.

    """

    BLOCKS = [
        Usina,
        DtbminAnos,
    ]



# Imports de módulos externos
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField


from inewave.newave.modelos.blocos.tabelacsv import TabelaCSV


class TabelaEcoEvap(TabelaCSV):
    """
    Bloco com o eco dos dados para representação linear
    da evaporação.
    """

    BEGIN_PATTERN = "-----;-----;"
    LINE_MODEL = Line(
        [
            IntegerField(size=5),
            IntegerField(size=5),
            LiteralField(size=14),
            FloatField(size=11, decimal_digits=2),
            FloatField(size=11, decimal_digits=2),
            IntegerField(size=8),
            IntegerField(size=11),
            IntegerField(size=11),
            IntegerField(size=15),
        ],
        delimiter=";",
    )
    COLUMN_NAMES = [
        "periodo",
        "codigo_usina",
        "nome_usina",
        "volume_referencia_hm3",
        "evaporacao_referencia_hm3",
        "coeficiente_evaporacao_mm_mes",
        "flag_evaporacao",
        "evaporacao_linear",
        "tipo_volume_referencia",
    ]
    END_PATTERN = ""



from cfinterface.files.sectionfile import SectionFile
from inewave.newave.modelos.cortesh import SecaoDadosCortesh

import pandas as pd  # type: ignore
from typing import TypeVar


class Cortesh(SectionFile):
    """
    Armazena os dados de saída do NEWAVE referentes ao
    cabeçalho dos cortes de Benders.
    """

    T = TypeVar("T")

    SECTIONS = [SecaoDadosCortesh]
    STORAGE = "BINARY"

    def __obtem_dados(self) -> SecaoDadosCortesh:
        dados = [r for r in self.data.of_type(SecaoDadosCortesh)]
        return dados[0]

    @property
    def versao_newave(self) -> int:
        """
        A versão do NEWAVE como um número inteiro (ex. 280000).

        :return: A versão do NEWAVE
        :rtype: int
        """
        return self.__obtem_dados().versao_newave

    @versao_newave.setter
    def versao_newave(self, v: int):
        self.__obtem_dados().versao_newave = v

    @property
    def tamanho_corte(self) -> int:
        """
        O tamanho (em bytes) do registro que contém os coeficientes
        de um corte.

        :return: O tamanho do corte
        :rtype: int
        """
        return self.__obtem_dados().tamanho_corte

    @tamanho_corte.setter
    def tamanho_corte(self, v: int):
        self.__obtem_dados().tamanho_corte = v

    @property
    def tamanho_estado(self) -> int:
        """
        O tamanho (em bytes) do registro que contém um estado
        visitado durante o cálculo da política.

        :return: O tamanho do estado
        :rtype: int
        """
        return self.__obtem_dados().tamanho_estado

    @tamanho_estado.setter
    def tamanho_estado(self, v: int):
        self.__obtem_dados().tamanho_estado = v

    @property
    def numero_rees(self) -> int:
        """
        O número de REEs existentes no caso

        :return: O número de REEs
        :rtype: int
        """
        return self.__obtem_dados().numero_rees

    @numero_rees.setter
    def numero_rees(self, v: int):
        self.__obtem_dados().numero_rees = v

    @property
    def numero_estagios_pre(self) -> int:
        """
        O número de estágios no período pré-estudo.

        :return: O número de estágios do período pré
        :rtype: int
        """
        return self.__obtem_dados().numero_estagios_pre

    @numero_estagios_pre.setter
    def numero_estagios_pre(self, v: int):
        self.__obtem_dados().numero_estagios_pre = v

    @property
    def numero_estagios_estudo(self) -> int:
        """
        O número de estágios no período de estudo.

        :return: O número de estágios do período de estudo.
        :rtype: int
        """
        return self.__obtem_dados().numero_estagios_estudo

    @numero_estagios_estudo.setter
    def numero_estagios_estudo(self, v: int):
        self.__obtem_dados().numero_estagios_estudo = v

    @property
    def numero_estagios_pos(self) -> int:
        """
        O número de estágios no período pós estudo.

        :return: O número de estágios do período pós
        :rtype: int
        """
        return self.__obtem_dados().numero_estagios_pos

    @numero_estagios_pos.setter
    def numero_estagios_pos(self, v: int):
        self.__obtem_dados().numero_estagios_pos = v

    @property
    def numero_estagios_ano(self) -> int:
        """
        O número de estágios existentes ao longo de
        um ano.

        :return: O número de estágios por ano
        :rtype: int
        """
        return self.__obtem_dados().numero_estagios_ano

    @numero_estagios_ano.setter
    def numero_estagios_ano(self, v: int):
        self.__obtem_dados().numero_estagios_ano = v

    @property
    def numero_configuracoes(self) -> int:
        """
        O número de configurações do sistema existentes no caso

        :return: O número de configurações
        :rtype: int
        """
        return self.__obtem_dados().numero_configuracoes

    @numero_configuracoes.setter
    def numero_configuracoes(self, v: int):
        self.__obtem_dados().numero_configuracoes = v

    @property
    def numero_forwards(self) -> int:
        """
        O número de séries forward utilizadas no cálculo
        da política.

        :return: O número de séries
        :rtype: int
        """
        return self.__obtem_dados().numero_forwards

    @numero_forwards.setter
    def numero_forwards(self, v: int):
        self.__obtem_dados().numero_forwards = v

    @property
    def numero_patamares(self) -> int:
        """
        O número de patamares de carga.

        :return: O número de patamares
        :rtype: int
        """
        return self.__obtem_dados().numero_patamares

    @numero_patamares.setter
    def numero_patamares(self, v: int):
        self.__obtem_dados().numero_patamares = v

    @property
    def ano_inicio_estudo(self) -> int:
        """
        O ano de início do período de estudo.

        :return: O ano de início
        :rtype: int
        """
        return self.__obtem_dados().ano_inicio_estudo

    @ano_inicio_estudo.setter
    def ano_inicio_estudo(self, v: int):
        self.__obtem_dados().ano_inicio_estudo = v

    @property
    def mes_inicio_estudo(self) -> int:
        """
        O mês de início do período de estudo. Considera a contagem
        sempre desde o Janeiro do primeiro ano.

        :return: O índice do mês de início.
        :rtype: int
        """
        return self.__obtem_dados().mes_inicio_estudo

    @mes_inicio_estudo.setter
    def mes_inicio_estudo(self, v: int):
        self.__obtem_dados().mes_inicio_estudo = v

    @property
    def lag_maximo_gnl(self) -> int:
        """
        O lag máximo para decisão do despacho antecipado
        das usinas térmicas GNL que é considerado.

        :return: O lag máximo em número de meses
        :rtype: int
        """
        return self.__obtem_dados().lag_maximo_gnl

    @lag_maximo_gnl.setter
    def lag_maximo_gnl(self, v: int):
        self.__obtem_dados().lag_maximo_gnl = v

    @property
    def numero_submercados(self) -> int:
        """
        Número de total de submercados, considerando apenas
        mercados não-fictícios.

        :return: O número de submercados
        :rtype: int
        """
        return self.__obtem_dados().numero_submercados

    @numero_submercados.setter
    def numero_submercados(self, v: int):
        self.__obtem_dados().numero_submercados = v

    @property
    def numero_total_submercados(self) -> int:
        """
        Número de total de submercados, considerando mercados
        fictícios e não-fictícios.

        :return: O número de submercados
        :rtype: int
        """
        return self.__obtem_dados().numero_total_submercados

    @numero_total_submercados.setter
    def numero_total_submercados(self, v: int):
        self.__obtem_dados().numero_total_submercados = v

    @property
    def mes_agregacao(self) -> int:
        """
        O mês em que se inicia a agregação em REEs em casos
        híbridos. Considera a contagem sempre a partir de janeiro
        do primeiro ano, mesmo se o período de estudo
        se iniciar em outro mês.

        :return: O mês
        :rtype: int
        """
        return self.__obtem_dados().mes_agregacao

    @mes_agregacao.setter
    def mes_agregacao(self, v: int):
        self.__obtem_dados().mes_agregacao = v

    @property
    def numero_maximo_uhes(self) -> int:
        """
        Número máximo de UHEs existentes em algum estágio
        do caso.

        :return: O número de UHEs.
        :rtype: int
        """
        return self.__obtem_dados().numero_maximo_uhes

    @numero_maximo_uhes.setter
    def numero_maximo_uhes(self, v: int):
        self.__obtem_dados().numero_maximo_uhes = v

    @property
    def considera_afluencia_anual(self) -> int:
        """
        Flag para indicação se é utilizado o modelo PAR(p)-A
        para geração de cenários.

        :return: O valor do flag
        :rtype: int
        """
        return self.__obtem_dados().considera_afluencia_anual

    @considera_afluencia_anual.setter
    def considera_afluencia_anual(self, v: int):
        self.__obtem_dados().considera_afluencia_anual = v

    @property
    def tipo_agregacao_caso(self) -> int:
        """
        Flag para indicação do tipo de agregação do caso,
        entre agregado, híbrido ou totalmente individualizado.

        :return: O valor do flag
        :rtype: int
        """
        return self.__obtem_dados().tipo_agregacao_caso

    @tipo_agregacao_caso.setter
    def tipo_agregacao_caso(self, v: int):
        self.__obtem_dados().tipo_agregacao_caso = v

    @property
    def estagio_individualizado_inicial(self) -> int:
        """
        O índice do estágio em que se inicia a simulação
        individualizada em UHE.

        :return: O índice do estágio
        :rtype: int
        """
        return self.__obtem_dados().estagio_individualizado_inicial

    @estagio_individualizado_inicial.setter
    def estagio_individualizado_inicial(self, v: int):
        self.__obtem_dados().estagio_individualizado_inicial = v

    @property
    def estagio_individualizado_final(self) -> int:
        """
        O índice do estágio em que termina a simulação
        individualizada em UHE.

        :return: O índice do estágio
        :rtype: int
        """
        return self.__obtem_dados().estagio_individualizado_final

    @estagio_individualizado_final.setter
    def estagio_individualizado_final(self, v: int):
        self.__obtem_dados().estagio_individualizado_final = v

    @property
    def tamanho_registro_individualizado(self) -> int:
        """
        O tamanho (em número de bytes) de um registro de
        corte para estágio individualizado.

        :return: O tamanho (em bytes)
        :rtype: int
        """
        return self.__obtem_dados().tamanho_registro_individualizado

    @tamanho_registro_individualizado.setter
    def tamanho_registro_individualizado(self, v: int):
        self.__obtem_dados().tamanho_registro_individualizado = v

    @property
    def estagio_agregado_inicial(self) -> int:
        """
        O índice do estágio em que se inicia a simulação
        agregada em REE.

        :return: O índice do estágio
        :rtype: int
        """
        return self.__obtem_dados().estagio_agregado_inicial

    @estagio_agregado_inicial.setter
    def estagio_agregado_inicial(self, v: int):
        self.__obtem_dados().estagio_agregado_inicial = v

    @property
    def estagio_agregado_final(self) -> int:
        """
        O índice do estágio em que termina a simulação
        agregada em REE.

        :return: O índice do estágio
        :rtype: int
        """
        return self.__obtem_dados().estagio_agregado_final

    @estagio_agregado_final.setter
    def estagio_agregado_final(self, v: int):
        self.__obtem_dados().estagio_agregado_final = v

    @property
    def tamanho_registro_agregado(self) -> int:
        """
        O tamanho (em número de bytes) de um registro de
        corte para estágio agregado.

        :return: O tamanho (em bytes)
        :rtype: int
        """
        return self.__obtem_dados().tamanho_registro_agregado

    @tamanho_registro_agregado.setter
    def tamanho_registro_agregado(self, v: int):
        self.__obtem_dados().tamanho_registro_agregado = v

    @property
    def ultimo_registro_cortes_estagio(self) -> pd.DataFrame:
        """
        Retorna os dados dos índices do último registro
        de cortes para cada estágio, para leitura do
        arquivo `cortes.dat`.

        - tipo_estagio [pre, estudo, pos] (`str`)
        - estagio (`int`)
        - indice_ultimo_corte (`int`)

        :return: Os dados dos índices dos cortes em uma tabela.
        :rtype: pd.DataFrame
        """
        return self.__obtem_dados().ultimo_registro_cortes_estagio

    @property
    def dados_submercados(self) -> pd.DataFrame:
        """
        Retorna os dados dos submercados e do número
        de REEs por submercados

        - codigo_submercado (`int`)
        - nome_submercado (`str`)
        - numero_rees_submercado (`int`)

        :return: Os dados dos submercados em uma tabela.
        :rtype: pd.DataFrame
        """
        return self.__obtem_dados().dados_submercados

    @property
    def dados_uhes(self) -> pd.DataFrame:
        """
        Retorna os dados das UHEs existentes no
        estudo realizado.

        - codigo_usina (`int`)
        - indice_usina (`int`)
        - posto (`int`)
        - ficticia (`int`)
        - codigo_submercado (`int`)
        - mes_agregacao (`int`)
        - codigo_interno_ree (`int`)
        - nome_ree (`str`)
        - codigo_submercado (`int`)
        - nome_submercado (`str`)

        :return: Os dados das UHEs em uma tabela.
        :rtype: pd.DataFrame
        """
        return self.__obtem_dados().dados_uhes



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.ghmaxsin import GHAnos


class Ghmaxsin(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes à geração hidraulica máxima
    por patamar para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `ghmaxsin.out`.
    """

    BLOCKS = [
        GHAnos,
    ]



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore
from inewave._utils.formatacao import (
    repete_vetor,
)
from inewave.config import MAX_REES, MESES_DF


class BlocoEafPast(Section):
    """
    Bloco de informações de vazões passadas
    por REE, existentes no arquivo `eafpast.dat`
    do NEWAVE.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [IntegerField(4, 0), LiteralField(10, 5)]
            + [FloatField(11, 15 + 11 * i, 2) for i in range(len(MESES_DF))]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoEafPast):
            return False
        bloco: BlocoEafPast = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(
                data={
                    "codigo_ree": repete_vetor(codigos_rees),
                    "nome_ree": repete_vetor(nomes_rees),
                    "mes": np.tile(np.arange(1, 13), len(nomes_rees)),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        i = 0
        tabela = np.zeros((MAX_REES, len(MESES_DF)))
        codigos_rees: List[int] = []
        nomes_rees: List[str] = []
        while True:
            linha = file.readline()
            # Confere se terminaram as usinas
            if len(linha) < 3:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            codigos_rees.append(dados[0])
            nomes_rees.append(dados[1])
            tabela[i, :] = dados[2:]
            i += 1

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do eafpast.dat não foram lidos com sucesso"
            )

        df = self.data.copy()
        rees = self.data["codigo_ree"].unique()
        for ree in rees:
            df_ree = df.loc[(df["codigo_ree"] == ree)]
            df_ree = df_ree.sort_values(["mes"])
            file.write(
                self.__linha.write(
                    [df_ree["codigo_ree"].iloc[0], df_ree["nome_ree"].iloc[0]]
                    + df_ree["valor"].tolist()
                )
            )



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousina import (
    ArquivoUsina,
)
from inewave.nwlistop.modelos.varmuh import VarmAnos


class Varmuh(ArquivoUsina):
    """
    Armazena os dados das saídas referentes aos armazenamentos por usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `varmuh00x.out`, onde x varia conforme a
    usina em questão.

    """

    BLOCKS = [
        Usina,
        VarmAnos,
    ]



# Imports de módulos externos
from cfinterface.components.block import Block
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
import pandas as pd  # type: ignore
from typing import IO, Dict, List


class TabelaAvlDesvFphaS(Block):
    """
    Bloco com os desvios da função de produção no plano de
    vazão vertida (S).
    """

    BEGIN_PATTERN = "-----;--------------;--------;"

    COLUMN_NAMES = [
        "codigo_usina",
        "nome_usina",
        "volume_armazenado_percentual",
        "vazao_turbinada_m3s",
        "vazao_vertida_m3s",
        "desvio_percentual",
    ]
    END_PATTERN = ""

    def _monta_df(self, dados: dict) -> pd.DataFrame:
        return pd.DataFrame(data=dados, columns=self.__class__.COLUMN_NAMES)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, TabelaAvlDesvFphaS):
            return False
        else:
            if not all(
                [type(self.data) is pd.DataFrame, type(o.data) is pd.DataFrame]
            ):
                return False
            else:
                return self.data.equals(o.data)

    def read(self, file: IO, *args, **kwargs):
        # Espera o fim do cabeçalho
        linha = file.readline()
        while True:
            linha = file.readline()
            if "VERT" in linha:
                num_vertimentos = linha.count("VERT")
                self.__linha = Line(
                    [
                        IntegerField(size=5),
                        LiteralField(size=14),
                        FloatField(size=8, decimal_digits=2),
                        FloatField(size=9, decimal_digits=2),
                    ]
                    + [
                        FloatField(size=9, decimal_digits=2)
                        for _ in range(num_vertimentos)
                    ],
                    delimiter=";",
                )
            if self.__class__.BEGIN_PATTERN in linha:
                break
            elif len(linha) < 3:
                return
        # Lê os valores das variáveis de um dos eixos
        dados_eixo = self.__linha.read(file.readline())
        valores_vert = dados_eixo[4:]
        num_valores_vert = len(valores_vert)
        # Ignora a segunda linha de cabeçalho
        file.readline()
        # Lê a tabela
        dados: Dict[str, List] = {c: [] for c in self.__class__.COLUMN_NAMES}
        while True:
            linha = file.readline()
            if len(linha) < 3 or "-----;--------------;------;" in linha:
                self.data = self._monta_df(dados)
                return
            dados_linha = self.__linha.read(linha)
            dados["codigo_usina"] += [dados_linha[0]] * num_valores_vert
            dados["nome_usina"] += [dados_linha[1]] * num_valores_vert
            dados["volume_armazenado_percentual"] += [
                dados_linha[2]
            ] * num_valores_vert
            dados["vazao_turbinada_m3s"] += [dados_linha[3]] * num_valores_vert
            dados["vazao_vertida_m3s"] += valores_vert
            dados["desvio_percentual"] += dados_linha[4:]



from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)

from inewave.nwlistop.modelos.edesvcsin import EdesvcsinAnos


class Edesvcsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes às energias
    de desvio de água controlável para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `edesvcsin.out`.

    """

    BLOCKS = [
        EdesvcsinAnos,
    ]



from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type, Optional
import pandas as pd  # type: ignore

from inewave.newave.modelos.penalid import BlocoPenalidades


class Penalid(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às penalidades
    aplicadas por desvio.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = [BlocoPenalidades]

    @property
    def penalidades(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as penalidades.

        - variavel (`str`)
        - codigo_ree_submercado (`int`)
        - patamar_penalidade (`int`)
        - patamar_carga (`int`)
        - valor_R$_MWh (`float`)
        - valor_R$_hm3 (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoPenalidades)
        if isinstance(b, BlocoPenalidades):
            return b.data
        return None

    @penalidades.setter
    def penalidades(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoPenalidades)
        if isinstance(b, BlocoPenalidades):
            b.data = df



"""
========================================
Análise do pmo.dat
========================================
"""


# %%
# Para realizar a análise do pmo.dat, será utilizado o módulo plotly
from inewave.newave import Pmo
import plotly.express as px
import plotly.io as pio

pio.templates.default = "ggplot2"


arq_pmo = Pmo.read("./newave/pmo.dat")

# %%
# Os dados associados ao processo de convergência são obtidos com:
arq_pmo.convergencia.head(10)

# %%
# O formato dos dados de custos por etapa também são acessíveis:
arq_pmo.custo_operacao_series_simuladas.head(10)

# %%
# Também é possível acessar o custo total:
arq_pmo.custo_operacao_total

# %%
# Cada arquivo pode ser visualizado de diferentes maneiras, a depender da aplicação.
# Por exemplo, é comum avaliar a convergência do modelo através da variação do Zinf.

fig = px.line(
    arq_pmo.convergencia,
    x="iteracao",
    y=["zinf", "zsup"],
)
fig

# %%
# Quando se analisam os custos de cada fonte, geralmente são feitos gráficos de barras
# empilhadas ou setores:

custos = arq_pmo.custo_operacao_series_simuladas
fig = px.pie(
    custos.loc[custos["valor_esperado"] > 0],
    values="valor_esperado",
    names="parcela",
)
fig

# %%
# Também é possível acessar diversas outras informações existentes no arquivo, como o
# relatório das produtibilidades calculadas e a tendência hidrológica vista pelo modelo.
# Mais informações estão disponíveis na página de referência.



from inewave.newave.modelos.selcor import (
    BlocoDadosSelcor,
)

from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type, Optional


class Selcor(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à metodologia
    de seleção de cortes.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = [
        BlocoDadosSelcor,
    ]

    @property
    def iteracao_inicial(self) -> Optional[int]:
        """
        O número da iteração inicial em que é aplicada a seleção
        de cortes.

        :return: O número da iteração
        :rtype: int | None
        """
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            return b.data[0][0]
        return None

    @iteracao_inicial.setter
    def iteracao_inicial(self, valor: int):
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            b.data[0][0] = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def tamanho_janela(self) -> Optional[int]:
        """
        O tamanho da janela de cortes ativos.

        :return: A janela
        :rtype: int | None
        """
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            return b.data[1][0]
        return None

    @tamanho_janela.setter
    def tamanho_janela(self, valor: int):
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            b.data[1][0] = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def numero_cortes_adicionados_por_iteracao(self) -> Optional[int]:
        """
        O número de cortes que são adicionados a cada iteração
        do processo de seleção de cortes.

        :return: O número de cortes adicionados.
        :rtype: int | None
        """
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            return b.data[2][0]
        return None

    @numero_cortes_adicionados_por_iteracao.setter
    def numero_cortes_adicionados_por_iteracao(self, valor: int):
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            b.data[2][0] = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def considera_cortes_da_propria_iteracao(self) -> Optional[int]:
        """
        O flag para considerar ou não os cortes construídos na própria
        iteração durante a backward.

        :return: O valor do flag
        :rtype: int | None
        """
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            return b.data[3][0]
        return None

    @considera_cortes_da_propria_iteracao.setter
    def considera_cortes_da_propria_iteracao(self, valor: int):
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            b.data[3][0] = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def imprime_relatorio(self) -> Optional[int]:
        """
        O flag para realizar ou não a impressão do relatório de
        saída do processo de seleção de cortes.

        :return: O valor do flag
        :rtype: int | None
        """
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            return b.data[4][0]
        return None

    @imprime_relatorio.setter
    def imprime_relatorio(self, valor: int):
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            b.data[4][0] = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def iteracoes_impressao(self) -> Optional[List[int]]:
        """
        O intervalo de iterações para as quais será realizada
        a impressão dos valores de saída.

        :return: As iterações inicial e final
        :rtype: list[int] | None
        """
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            return b.data[5]
        return None

    @iteracoes_impressao.setter
    def iteracoes_impressao(self, valor: List[int]):
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            b.data[5] = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def series_impressao(self) -> Optional[List[int]]:
        """
        O intervalo de séries para as quais será realizada
        a impressão dos valores de saída.

        :return: As séries inicial e final
        :rtype: list[int] | None
        """
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            return b.data[6]
        return None

    @series_impressao.setter
    def series_impressao(self, valor: List[int]):
        b = self.data.get_sections_of_type(BlocoDadosSelcor)
        if isinstance(b, BlocoDadosSelcor):
            b.data[6] = valor
        else:
            raise ValueError("Campo não lido")



from inewave.config import MESES_DF

from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField

from inewave.nwlistop.modelos.blocos.valoresseriepatamar import (
    ValoresSeriePatamar,
)


class VdesviouhAnos(ValoresSeriePatamar):
    """
    Bloco com as informações das tabelas de volume desviado de água
    da usina por mês/ano de estudo.
    """

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [  # type: ignore
            IntegerField(4, 2),
            LiteralField(5, 6),
        ]
        + [FloatField(9, 14 + 9 * i, 2) for i in range(len(MESES_DF))]  # type: ignore
    )



from inewave.newave.modelos.parpvaz import BlocoSerieVazoesUHE
from inewave.newave.modelos.parpvaz import BlocoCorrelVazoesUHE
from inewave.newave.modelos.parpvaz import BlocoCorrelParcialVazoesUHE
from inewave.newave.modelos.parpvaz import BlocoOrdemModeloUHE
from inewave.newave.modelos.parpvaz import BlocoCoeficientesModeloUHE
from inewave.newave.modelos.parpvaz import BlocoSerieRuidosUHE
from inewave.newave.modelos.parpvaz import BlocoCorrelRuidosUHE
from inewave.newave.modelos.parpvaz import BlocoCorrelEspacialAnualMensalUHE

from cfinterface.components.block import Block
from cfinterface.files.blockfile import BlockFile
from typing import Type, TypeVar, Optional, Any, List
import pandas as pd  # type: ignore
from datetime import datetime


class Parpvaz(BlockFile):
    """
    Armazena os dados de saída do NEWAVE referentes aos modelos e às
    séries sintéticas de vazões geradas pelo PAR(p).


    Esta classe lida com informações de saída do NEWAVE e
    cujas saídas devem ser compatíveis com as observadas através
    do NWLISTOP.

    """

    T = TypeVar("T")

    BLOCKS = [
        BlocoSerieVazoesUHE,
        BlocoCorrelVazoesUHE,
        BlocoCorrelParcialVazoesUHE,
        BlocoOrdemModeloUHE,
        BlocoCoeficientesModeloUHE,
        BlocoSerieRuidosUHE,
        BlocoCorrelRuidosUHE,
        BlocoCorrelEspacialAnualMensalUHE,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)
        self.__series_vazoes = None
        self.__correl_series_vazoes = None
        self.__correl_parcial_series_vazoes = None
        self.__series_ruido = None
        self.__correl_series_ruido = None
        self.__ordem_original_modelo = None
        self.__ordem_final_modelo = None
        self.__coeficientes = None
        self.__correl_espacial_anual_mensal = None

    def __uhes(self) -> Optional[List[str]]:
        """
        Retorna a lista das UHEs lidos do arquivo.

        :return: Os nomes das UHEs
        :rtype: List[str]
        """
        if self.series_vazoes_uhe is None:
            return None
        else:
            return self.series_vazoes_uhe["uhe"].unique().tolist()

    def __concatena_dados(self, bloco: Type[Block]) -> Optional[Any]:
        """
        Obtém os dados de um bloco se este existir dentre os blocos do arquivo.

        :param bloco: O tipo do bloco cujos dados serão extraídos
        :type bloco: Type[T]
        :param indice: Qual dos blocos do tipo será acessado
        :type indice: int, optional
        :return: Os dados do bloco, se existirem
        :rtype: Any
        """
        dados = pd.DataFrame()
        for b in self.data.of_type(bloco):
            if dados.empty:
                dados = b.data
            else:
                dados = pd.concat([dados, b.data], ignore_index=True)
        if not dados.empty:
            return dados
        else:
            return None

    def __adiciona_coluna_uhe(
        self, df: Optional[pd.DataFrame]
    ) -> Optional[pd.DataFrame]:
        """
        Adiciona uma coluna com as UHEs de cada amostra, assumindo
        a mesma ordem das séries de vazões.

        :param df: O DataFrame que irá receber as UHEs
        :type df: pd.DataFrame
        :return: O DataFrame com as UHEs
        :rtype: pd.DataFrame
        """
        if df is None:
            return None
        uhes = self.__uhes()
        if uhes is None:
            return None
        linhas_por_uhe = df.shape[0] / len(uhes)
        if int(linhas_por_uhe) != linhas_por_uhe:
            raise ValueError(
                f"{df.shape[0]} linhas não podem ser "
                + f"divididas em {len(uhes)} grupos"
            )
        cols = list(df.columns)
        col_uhe: List[str] = []
        for uhe in uhes:
            col_uhe += [uhe] * int(linhas_por_uhe)
        df["uhe"] = col_uhe
        return df[["uhe"] + cols]

    def __adiciona_coluna_uhe_com_estagios(
        self, df: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Adiciona uma coluna com as UHEs de cada amostra e outra
        com o estágio de cada uma, assumindo
        a mesma ordem das séries de vazões.

        :param df: O DataFrame que irá receber as UHEs
        :type df: pd.DataFrame
        :return: O DataFrame com as UHEs
        :rtype: pd.DataFrame
        """
        uhes = self.__uhes()
        if uhes is None:
            return None
        linhas_por_uhe = df.shape[0] / len(uhes)
        if int(linhas_por_uhe) != linhas_por_uhe:
            raise ValueError(
                f"{df.shape[0]} linhas não podem ser "
                + f"divididas em {len(uhes)} grupos"
            )
        cols = list(df.columns)
        col_uhe: List[str] = []
        col_estagio: List[int] = []
        for uhe in uhes:
            col_uhe += [uhe] * int(linhas_por_uhe)
            col_estagio += list(range(1, int(linhas_por_uhe) + 1))
        df["uhe"] = col_uhe
        df["estagio"] = col_estagio
        return df[["uhe", "estagio"] + cols]

    def __adiciona_coluna_uhe_corrigindo_pre_pos(
        self, df: Optional[pd.DataFrame]
    ) -> Optional[pd.DataFrame]:
        """
        Adiciona uma coluna com as UHEs de cada amostra e outra
        com o estágio de cada uma, assumindo
        a mesma ordem das séries de vazões, e corrigindo os valores
        dos anos se houve períodos PRE e POS.

        :param df: O DataFrame que irá receber as UHEs
        :type df: pd.DataFrame
        :return: O DataFrame com as UHEs
        :rtype: pd.DataFrame
        """

        def converte_vetor_anos(anos: List[str], n: int) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"]) // n
            numero_anos_pos = len([p for p in anos if p == "POS"]) // n
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # TODO - remover fallback de 1970 quando estiver o ano correto.
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = (
                sorted(anos_estudo)[0] if len(anos_estudo) > 0 else 1970
            )
            # Descobre o último ano de estudo
            ultimo_ano_estudo = (
                sorted(anos_estudo)[-1] if len(anos_estudo) > 0 else 1970
            )
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx_i = n * a
                idx_f = idx_i + n
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx_i:idx_f] = [str(ano)] * n
            for a in range(numero_anos_pos):
                idx_i = indice_inicio_pos + n * a
                idx_f = idx_i + n
                ano = ultimo_ano_estudo + a + 1
                anos[idx_i:idx_f] = [str(ano)] * n
            return [int(a) for a in anos]

        if df is None:
            return None
        uhes = self.__uhes()
        if uhes is None:
            return None
        linhas_por_uhe = df.shape[0] / len(uhes)
        if int(linhas_por_uhe) != linhas_por_uhe:
            raise ValueError(
                f"{df.shape[0]} linhas não podem ser "
                + f"divididas em {len(uhes)} grupos"
            )
        cols = list(df.columns)
        col_uhe: List[str] = []
        for uhe in uhes:
            col_uhe += [uhe] * int(linhas_por_uhe)
        df["uhe"] = col_uhe
        uhe0 = uhes[0]
        ano0 = df["ano"].unique().tolist()[0]
        filtro = (df["ano"] == ano0) & (df["uhe"] == uhe0)
        n_series = df.loc[filtro].shape[0]
        for i, uhe in enumerate(uhes):
            i_i = i * int(linhas_por_uhe)
            i_f = i_i + int(linhas_por_uhe) - 1
            df.loc[i_i:i_f, "ano"] = converte_vetor_anos(
                df.loc[i_i:i_f, "ano"].tolist(), n_series
            )
        return df[["uhe"] + cols]

    def __converte_ano_mes_data(
        self, df: Optional[pd.DataFrame]
    ) -> Optional[pd.DataFrame]:
        """
        Converte um dataframe com colunas `mes` e `ano` para um com
        uma coluna `data`.

        :param df: O DataFrame com colunas `mes` e `ano`
        :type df: pd.DataFrame
        :return: O DataFrame com coluna `data`
        :rtype: pd.DataFrame
        """
        if df is None:
            return None
        cols_identificacao = [
            c for c in df.columns if c not in ["valor", "ano", "mes"]
        ]
        df = df.copy()
        df.loc[:, "data"] = df.apply(
            lambda linha: datetime(
                year=linha["ano"], month=linha["mes"], day=1
            ),
            axis=1,
        )
        return df.drop(columns=["ano", "mes"])[
            cols_identificacao + ["data", "valor"]
        ]

    @property
    def series_vazoes_uhe(self) -> Optional[pd.DataFrame]:
        """
        A tabela de séries de vazões para todas as configurações
        e UHEs, no mesmo formato do arquivo `parpvaz.dat`.

        - uhe (`str`)
        - configuracao (`int`)
        - data (`datetime`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        if self.__series_vazoes is None:
            self.__series_vazoes = self.__concatena_dados(BlocoSerieVazoesUHE)
        return self.__series_vazoes

    @property
    def series_ruido_uhe(self) -> Optional[pd.DataFrame]:
        """
        A tabela de séries de ruído para todas as UHEs,
        no mesmo formato do arquivo `parpvaz.dat`.

        - uhe (`str`)
        - serie (`int`)
        - data (`datetime`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        if self.__series_ruido is None:
            self.__series_ruido = self.__concatena_dados(BlocoSerieRuidosUHE)
            self.__series_ruido = (
                self.__adiciona_coluna_uhe_corrigindo_pre_pos(
                    self.__series_ruido
                )
            )
            self.__series_ruido = self.__converte_ano_mes_data(
                self.__series_ruido
            )
        return self.__series_ruido

    @property
    def correlacao_series_vazoes_uhe(self) -> Optional[pd.DataFrame]:
        """
        A tabela de correlação das séries de vazões para
        todas as configurações vigentes e UHEs,
        no mesmo formato do arquivo `parpvaz.dat`.

        - uhe (`str`)
        - data (`datetime`)
        - lag (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        if self.__correl_series_vazoes is None:
            self.__correl_series_vazoes = self.__concatena_dados(
                BlocoCorrelVazoesUHE
            )
            self.__correl_series_vazoes = self.__adiciona_coluna_uhe(
                self.__correl_series_vazoes
            )
        return self.__correl_series_vazoes

    @property
    def correlacao_parcial_series_vazoes_uhe(self) -> Optional[pd.DataFrame]:
        """
        A tabela de correlação das séries de vazões para
        todas as configurações vigentes e UHEs,
        no mesmo formato do arquivo `parpvaz.dat`.

        - uhe (`str`)
        - data (`datetime`)
        - lag (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        if self.__correl_parcial_series_vazoes is None:
            self.__correl_parcial_series_vazoes = self.__concatena_dados(
                BlocoCorrelParcialVazoesUHE
            )
            self.__correl_parcial_series_vazoes = self.__adiciona_coluna_uhe(
                self.__correl_parcial_series_vazoes
            )
        return self.__correl_parcial_series_vazoes

    @property
    def correlacao_series_ruidos_uhe(self) -> Optional[pd.DataFrame]:
        """
        A tabela de correlação das séries de ruídos para
        todas as configurações vigentes e UHEs,
        no mesmo formato do arquivo `parpvaz.dat`.

        - uhe (`str`)
        - data (`datetime`)
        - lag (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        if self.__correl_series_ruido is None:
            self.__correl_series_ruido = self.__concatena_dados(
                BlocoCorrelRuidosUHE
            )
            self.__correl_series_ruido = self.__adiciona_coluna_uhe(
                self.__correl_series_ruido
            )
        return self.__correl_series_ruido

    @property
    def ordem_original_modelo(self) -> Optional[pd.DataFrame]:
        """
        A tabela de ordens originais do modelo PAR ou PAR-A
        de cada UHE, no mesmo formato do arquivo `parpvaz.dat`,
        organizada por ano de estudo.

        - uhe (`str`)
        - data (`datetime`)
        - valor (`int`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """

        if self.__ordem_original_modelo is None:
            dados = self.__concatena_dados(BlocoOrdemModeloUHE)
            if dados is not None:
                dados = (
                    dados.loc[dados["tipo"] == "ORIGINAL", :]
                    .drop(columns=["tipo"])
                    .copy()
                )
                dados = self.__adiciona_coluna_uhe(dados)
                self.__ordem_original_modelo = dados

        return self.__ordem_original_modelo

    @property
    def ordem_final_modelo(self) -> Optional[pd.DataFrame]:
        """
        A tabela de ordens finais do modelo PAR ou PAR-A
        de cada UHE, no mesmo formato do arquivo `parpvaz.dat`,
        organizada por ano de estudo.

        - uhe (`str`)
        - data (`datetime`)
        - valor (`int`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        if self.__ordem_final_modelo is None:
            dados = self.__concatena_dados(BlocoOrdemModeloUHE)
            if dados is not None:
                dados = (
                    dados.loc[dados["tipo"] == "FINAL", :]
                    .drop(columns=["tipo"])
                    .copy()
                )
                dados = self.__adiciona_coluna_uhe(dados)
                self.__ordem_final_modelo = dados

        return self.__ordem_final_modelo

    @property
    def coeficientes(self) -> Optional[pd.DataFrame]:
        """
        Lista de coeficientes dos modelos PAR ou PAR-A
        de cada UHE, no mesmo formato do arquivo `parpvaz.dat`,
        organizada por período de estudo.

        - uhe (`str`)
        - estagio (`int`)
        - tipo (`str`)
        - ordem (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        if self.__coeficientes is None:
            dados = self.__concatena_dados(BlocoCoeficientesModeloUHE)
            if dados is not None:
                dados = self.__adiciona_coluna_uhe_com_estagios(dados)
                self.__coeficientes = dados

        return self.__coeficientes

    @property
    def correlacao_espacial_anual_mensal(self) -> Optional[pd.DataFrame]:
        """
        A tabela de correlação para todas as UHEs,
        no mesmo formato do arquivo `parpvaz.dat`.

        - uhe_1 (`str`)
        - uhe_2 (`str`)
        - janeiro (`str`)
        - fevereiro (`str`)
        - ...
        - dezembro (`str`)
        - anual (`str`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        if self.__correl_espacial_anual_mensal is None:
            self.__correl_espacial_anual_mensal = self.__concatena_dados(
                BlocoCorrelEspacialAnualMensalUHE
            )
        return self.__correl_espacial_anual_mensal
